<!doctype html><html lang="en-us"><head><meta charset="utf-8"></head>
<body>
<canvas style='display:block; margin:auto;'></canvas>
<script>

#if !MODULARIZE
  var {{{ EXPORT_NAME }}} = {};
#endif

#if WASM == 2
  var supportsWasm = window.WebAssembly && location.search.indexOf('_rwasm=0') < 0;
#endif

  // Depending on the build flags that one uses, different files need to be downloaded
  // to load the compiled page. The right set of files will be expanded to be downloaded
  // via the directive below.
{{{ DOWNLOAD_JS_AND_WASM_FILES }}}

#if SINGLE_FILE
// If we are doing a SINGLE_FILE=1 build, inlined JS runtime code follows here:
{{{ JS_CONTENTS_IN_SINGLE_FILE_BUILD }}}
#endif

</script>
</body>
</html>
