#define SYS_chdir		 __syscall_chdir
#define SYS_chmod		 __syscall_chmod
#define SYS_getpid		 __syscall_getpid
#define SYS_sync		 __syscall_sync
#define SYS_rmdir		 __syscall_rmdir
#define SYS_dup		 __syscall_dup
#define SYS_acct		 __syscall_acct
#define SYS_ioctl		 __syscall_ioctl
#define SYS_setpgid		 __syscall_setpgid
#define SYS_umask		 __syscall_umask
#define SYS_getppid		 __syscall_getppid
#define SYS_setsid		 __syscall_setsid
#define SYS_getrusage		 __syscall_getrusage
#define SYS_munmap		 __syscall_munmap
#define SYS_fchmod		 __syscall_fchmod
#define SYS_getpriority	 __syscall_getpriority
#define SYS_setpriority	 __syscall_setpriority
#define SYS_wait4		__syscall_wait4
#define SYS_setdomainname	__syscall_setdomainname
#define SYS_uname		__syscall_uname
#define SYS_mprotect		__syscall_mprotect
#define SYS_getpgid		__syscall_getpgid
#define SYS_fchdir		__syscall_fchdir
#define SYS_msync		__syscall_msync
#define SYS_getsid		__syscall_getsid
#define SYS_fdatasync		__syscall_fdatasync
#define SYS_mlock		__syscall_mlock
#define SYS_munlock		__syscall_munlock
#define SYS_mlockall		__syscall_mlockall
#define SYS_munlockall		__syscall_munlockall
#define SYS_mremap		__syscall_mremap
#define SYS_poll		__syscall_poll
#define SYS_getcwd		__syscall_getcwd
#define SYS_mmap2		__syscall_mmap2
#define SYS_truncate64		__syscall_truncate64
#define SYS_ftruncate64	__syscall_ftruncate64
#define SYS_stat64		__syscall_stat64
#define SYS_lstat64		__syscall_lstat64
#define SYS_fstat64		__syscall_fstat64
#define SYS_getuid32		__syscall_getuid32
#define SYS_getgid32		__syscall_getgid32
#define SYS_geteuid32		__syscall_geteuid32
#define SYS_getegid32		__syscall_getegid32
#define SYS_getgroups32	__syscall_getgroups32
#define SYS_fchown32		__syscall_fchown32
#define SYS_getresuid32	__syscall_getresuid32
#define SYS_getresgid32	__syscall_getresgid32
#define SYS_mincore		__syscall_mincore
#define SYS_madvise		__syscall_madvise
#define SYS_getdents64		__syscall_getdents64
#define SYS_fcntl64		__syscall_fcntl64
#define SYS_statfs64		__syscall_statfs64
#define SYS_fstatfs64		__syscall_fstatfs64
#define SYS_fadvise64		__syscall_fadvise64
#define SYS_openat		__syscall_openat
#define SYS_mkdirat		__syscall_mkdirat
#define SYS_mknodat		__syscall_mknodat
#define SYS_fchownat		__syscall_fchownat
#define SYS_newfstatat		__syscall_newfstatat
#define SYS_unlinkat		__syscall_unlinkat
#define SYS_renameat		__syscall_renameat
#define SYS_linkat		__syscall_linkat
#define SYS_symlinkat		__syscall_symlinkat
#define SYS_readlinkat		__syscall_readlinkat
#define SYS_fchmodat2		__syscall_fchmodat2
#define SYS_faccessat		__syscall_faccessat
#define SYS_utimensat		__syscall_utimensat
#define SYS_fallocate		__syscall_fallocate
#define SYS_dup3		__syscall_dup3
#define SYS_pipe2		__syscall_pipe2
#define SYS_recvmmsg		__syscall_recvmmsg
#define SYS_prlimit64		__syscall_prlimit64
#define SYS_sendmmsg		__syscall_sendmmsg
#define SYS_socket		__syscall_socket
#define SYS_socketpair		__syscall_socketpair
#define SYS_bind		__syscall_bind
#define SYS_connect		__syscall_connect
#define SYS_listen		__syscall_listen
#define SYS_accept4		__syscall_accept4
#define SYS_getsockopt		__syscall_getsockopt
#define SYS_setsockopt		__syscall_setsockopt
#define SYS_getsockname		__syscall_getsockname
#define SYS_getpeername		__syscall_getpeername
#define SYS_sendto		__syscall_sendto
#define SYS_sendmsg		__syscall_sendmsg
#define SYS_recvfrom		__syscall_recvfrom
#define SYS_recvmsg		__syscall_recvmsg
#define SYS_shutdown		__syscall_shutdown
