#ifndef LLVM_LIBC_MACROS_LINUX_ERROR_NUMBER_MACROS_H
#define LLVM_LIBC_MACROS_LINUX_ERROR_NUMBER_MACROS_H

#ifndef ECANCELED
#define ECANCELED 125
#endif // ECANCELED

#ifndef EOWNERDEAD
#define EOWNERDEAD 130
#endif // EOWNERDEAD

#ifndef ENOTRECOVERABLE
#define ENOTRECOVERABLE 131
#endif // ENOTRECOVERABLE

#ifndef ERFKILL
#define ERFKILL 132
#endif // ERFKILL

#ifndef EHWPOISON
#define EHWPOISON 133
#endif // EHWPOISON

#ifndef EOPNOTSUPP
#define EOPNOTSUPP 95
#endif

#ifndef ENOTSUP
#define ENOTSUP EOPNOTSUPP
#endif

#endif // LLVM_LIBC_MACROS_LINUX_ERROR_NUMBER_MACROS_H
