//===-- Definition of macros to be used with file seek functions ----------===//
//
// Part of the LLVM Project, under the Apache License v2.0 with LLVM Exceptions.
// See https://llvm.org/LICENSE.txt for license information.
// SPDX-License-Identifier: Apache-2.0 WITH LLVM-exception
//
//===----------------------------------------------------------------------===//

#ifndef LLVM_LIBC_MACROS_FILE_SEEK_MACROS_H
#define LLVM_LIBC_MACROS_FILE_SEEK_MACROS_H

#define SEEK_SET 0
#define SEEK_CUR 1
#define SEEK_END 2

#endif // LLVM_LIBC_MACROS_FILE_SEEK_MACROS_H
