// Copyright 2021 The Emscripten Authors.  All rights reserved.
// Emscripten is available under two separate licenses, the MIT license and the
// University of Illinois/NCSA Open Source License.  Both these licenses can be
// found in the LICENSE file.

// stb_image 2.x need to have STB_IMAGE_IMPLEMENTATION defined to include
// the implementation when compiling
#define STB_IMAGE_IMPLEMENTATION
#include <stb_image.h>
