//===-- Definition of EFI_VIRTUAL_ADDRESS type ----------------------------===//
//
// Part of the LLVM Project, under the Apache License v2.0 with LLVM Exceptions.
// See https://llvm.org/LICENSE.txt for license information.
// SPDX-License-Identifier: Apache-2.0 WITH LLVM-exception
//
//===----------------------------------------------------------------------===//

#ifndef LLVM_LIBC_TYPES_EFI_VIRTUAL_ADDRESS_H
#define LLVM_LIBC_TYPES_EFI_VIRTUAL_ADDRESS_H

#include "../llvm-libc-macros/stdint-macros.h"

typedef uint64_t EFI_VIRTUAL_ADDRESS;

#endif // LLVM_LIBC_TYPES_EFI_VIRTUAL_ADDRESS_H
