// Typical input to minifyGlobals, generated by wasm2js.

function instantiate(wasmImports) {
 function Table(ret) {
  ret.set = function(i, func) {
   this[i] = func;
  };
  ret.get = function(i) {
   return this[i];
  };
  return ret;
 }
 function asmFunc(env) {
  var memory = env.a;
  var buffer = memory.buffer;
  var HEAP8 = new Int8Array(buffer);
  var HEAP16 = new Int16Array(buffer);
  var HEAP32 = new Int32Array(buffer);
  var HEAPU8 = new Uint8Array(buffer);
  var HEAPU16 = new Uint16Array(buffer);
  var HEAPU32 = new Uint32Array(buffer);
  var HEAPF32 = new Float32Array(buffer);
  var HEAPF64 = new Float64Array(buffer);
  var Math_imul = Math.imul;
  var Math_fround = Math.fround;
  var Math_abs = Math.abs;
  var Math_clz32 = Math.clz32;
  var Math_min = Math.min;
  var Math_max = Math.max;
  var Math_floor = Math.floor;
  var Math_ceil = Math.ceil;
  var Math_trunc = Math.trunc;
  var Math_sqrt = Math.sqrt;
  var abort = env.abort;
  var nan = NaN;
  var infinity = Infinity;
  var fimport$0 = env.b;
  var fimport$1 = env.c;
  var global$0 = 5245136;
  EMSCRIPTEN_FUNCS();
  var FUNCTION_TABLE = Table([ null, $6, $7, $5 ]);
  function __wasm_memory_size() {
   return buffer.byteLength / 65536 | 0;
  }
  return {
   "d": $2,
   "e": $11,
   "f": FUNCTION_TABLE
  };
 }
 return asmFunc(wasmImports);
}
// EXTRA_INFO:{"globals": ["$0", "$1", "$2", "$3", "$4", "$5", "$6", "$7", "$8", "$9", "$10", "$11"]}
