// Copyright 2019 The Emscripten Authors.  All rights reserved.
// Emscripten is available under two separate licenses, the MIT license and the
// University of Illinois/NCSA Open Source License.  Both these licenses can be
// found in the LICENSE file.

#include <stdio.h>
#include <emscripten.h>

int main() {
  int x = EM_ASM_INT({
    globalThis.disableErrorReporting = true;
    var checkError = (err) => {
      if (!err) return;
      var message = err.toString();
      var success = message.indexOf("unreachable") >= 0 || // firefox
                    message.indexOf("Script error.") >= 0; // chrome
      if (success && !Module.reported) {
        Module.reported = true;
        console.log("reporting success");
        // manually REPORT_RESULT; we shouldn't call back into native code at this point
        reportResultToServer(0);
      }
    };
    window.onerror = (message, source, lineno, colno, error) => {
      checkError(error || message);
    };
    window.onunhandledrejection = (e) => {
      checkError(e.reason);
    };
    return 0;
  });

  emscripten_sleep(1);

  // We should not get here - the unwind will fail as we did now all the right
  // functions - this function should be instrumented, but will not be.
  puts("We should not get here!");
  REPORT_RESULT(1);
  return 1;
}
