//===-- Definition of EFI_ALLOCATE_TYPE type ------------------------------===//
//
// Part of the LLVM Project, under the Apache License v2.0 with LLVM Exceptions.
// See https://llvm.org/LICENSE.txt for license information.
// SPDX-License-Identifier: Apache-2.0 WITH LLVM-exception
//
//===----------------------------------------------------------------------===//

#ifndef LLVM_LIBC_TYPES_EFI_ALLOCATE_TYPE_H
#define LLVM_LIBC_TYPES_EFI_ALLOCATE_TYPE_H

typedef enum {
  AllocateAnyPages,
  AllocateMaxAddress,
  AllocateAddress,
  MaxAllocateType
} EFI_ALLOCATE_TYPE;

#endif // LLVM_LIBC_TYPES_EFI_ALLOCATE_TYPE_H
