/*
 * Copyright 2019 The Emscripten Authors.  All rights reserved.
 * Emscripten is available under two separate licenses, the MIT license and the
 * University of Illinois/NCSA Open Source License.  Both these licenses can be
 * found in the LICENSE file.
 */

// This header has been moved to the LLVM project:
// https://github.com/llvm/llvm-project/blob/main/clang/lib/Headers/wasm_simd128.h
#include_next <wasm_simd128.h>
