//===-- Definition of the offsetof macro ----------------------------------===//
//
// Part of the LLVM Project, under the Apache License v2.0 with LLVM Exceptions.
// See https://llvm.org/LICENSE.txt for license information.
// SPDX-License-Identifier: Apache-2.0 WITH LLVM-exception
//
//===----------------------------------------------------------------------===//

#ifndef LLVM_LIBC_MACROS_OFFSETOF_MACRO_H
#define LLVM_LIBC_MACROS_OFFSETOF_MACRO_H

#define __need_offsetof
#include <stddef.h>

#endif // LLVM_LIBC_MACROS_OFFSETOF_MACRO_H
